;Tetrade Huffman compress/decompress functions.
;Calling example:
;   stdcall HufCompressTet, packing_array, packing_array_size, packed_array, 1
;   stdcall HufDecompressTet, packed_array, buffer_for_unpacking, 1

;in: lpSrc - pointer to packed data
;    lpRes - pointer to result buffer
;    dwCallMode - 0 - get required size of result buffer
;                 1 - decompress data
;out:
;    dwCallMode:    0           1
;failed             0           0
;successfull  required size    true
;base independense foo. size < 0E0h bytes
proc HufDecompressTet lpSrc:DWORD, lpRes:DWORD, dwCallMode:DWORD
local buffer[14h]:DWORD, pRes:DWORD
    cmp [dwCallMode], 0
    jne @f
    mov esi, [lpSrc]
    mov eax, [esi+4]
    ret
@@:
    mov esi, [lpSrc]
    lea edi, [buffer]
    mov ecx, 14h
    rep movsd

    mov esi, [lpSrc]
    mov edi, [lpRes]
    mov [pRes], edi

    mov ecx, [esi+4]
    xor eax, eax
    rep stosb

    xor ecx, ecx
    xor ebx, ebx
    mov edi, 50h

    mov dl, [esi+edi]
    ror edx, 8
    inc edi
.next_loop:
    sub dword [esi+0Ch], 1
    jnc @f
    cmp dword [esi+8], 0
    jz .unpacked
    dec dword [esi+8]
@@:
    cmp ch, 8h
    jb @f
    cmp dword [esi+0Ch], 0
    ja .get_nxt_byte
    test dword [esi+8], 0
    je @f
.get_nxt_byte:
    sub ch, 8
    rol edx, 8
    mov dl, [esi+edi]
    ror edx, 8
    inc edi
@@:
    inc cl	     ; 
    inc ch	     ;  edx
    rol edx, 1
    or eax, -1
    shl eax, cl
    not eax
    and eax, edx

    push ebx
    mov ebx, 10h
@@:
    cmp [esi+ebx*4+0Dh], cl
    jne .wrong
    cmp [esi+ebx*4+0Eh], ax
    jne .wrong
    mov al, [esi+ebx*4+0Ch]
    xor cl, cl
    pop ebx
    jmp .checked
.wrong:
    dec ebx
    jnz @b
    pop ebx

.checked:
    test cl, cl
    jnz .next_loop
    test bh, bh
    jnz @f
    shl al, 4
    mov bl, al
    mov bh, 1
    jmp .next_loop
@@:
    and al, 0Fh
    or bl, al
    push edi
    mov edi, [pRes]
    mov [edi], bl
    xor ebx, ebx
    inc [pRes]
    pop edi
    jmp .next_loop
.unpacked:
    lea esi, [buffer]
    mov edi, [lpSrc]
    mov ecx, 14h
    rep movsd
    mov eax, ebp
    ret
.fexit:
    xor eax, eax
    ret
endp







;in: lpSrc - pointer to source
;    dwSizeOfSrc - size of source
;    lpRes - pointer to result buffer
;    dwCallMode - 0 - get required size of result buffer,
;                 1 - compress data
;out:
;    dwCallMode   0                1
;failed:          0                0
;successfull  required size      true
proc HufCompressTet lpSrc:DWORD, dwSizeOfSrc:DWORD, lpRes:DWORD, dwCallMode:DWORD
local weights[16]:DWORD, nodes[200h]:DWORD, leafs[32]:DWORD, topnodes[16]:DWORD,\
      udata[16]:DWORD, NumberOfLeafs:DWORD, LNumberOfBits:DWORD, HNumberOfBits:DWORD,\
      SizeOfRes:DWORD
;leafs[] format: tetrade0:dword, weight0:dword, tetrade1:dword,...
;nodes[] format: weight:dword, pLNode:dword, pRNode:dword, pParent:dword
;udata[] format: tetrade:byte, hufcode_length:byte, hufcode(   ):word
;  :
;AlgoID:dword  - id   (  0x00000000)
;SizeBefore:dword -    
;HNumberOfBits:dword -       (RAW data) ( )
;LNumberOfBits:dword -       (RAW data) ( )
;UData[10h]:dword    -    (tetrade0:byte, code_length0:byte, code:word;tetrad1:byte,...)
;   50h 
    pushad
    cmp [dwSizeOfSrc], 80000000h
    ja .fexit
    cmp [dwSizeOfSrc], 0
    je .fexit

    xor eax, eax
    lea edi, [weights]
    mov ecx, 16
    cld
    rep stosd

    lea edi, [nodes]
    mov ecx, 200h
    rep stosd

    lea edi, [leafs]
    mov ecx, 32
    rep stosd

    lea edi, [udata]
    mov ecx, 16
    rep stosd

    call .calc_weights
    call .init_leafs
    mov [NumberOfLeafs], eax
    call .grow_tree
    call .init_udata
    call .calc_required_size
    mov [SizeOfRes], eax
    add [SizeOfRes], 50h
    cmp [dwCallMode], 0
    jnz @f
    popad
    mov eax, [SizeOfRes]
    ret
@@:
    call .pack_data
    popad
    mov eax, ebp
    ret
.fexit:
    popad
    xor eax, eax
    ret

.calc_weights:
    xor ebx, ebx
    lea edi, [weights]
    mov ecx, [dwSizeOfSrc]
    mov esi, [lpSrc]
@@:
    mov al, byte [esi]
    mov bl, al
    and bl, 0Fh
    inc dword[edi+ebx*4]
    mov bl, al
    shr bl, 4
    inc dword[edi+ebx*4]
    inc esi
    loop @b
@@:
    retn

.init_leafs:
    xor edx, edx
    lea esi, [weights]
    lea edi, [leafs]
    xor ecx, ecx
.init_next_leaf:
    mov eax, [esi+ecx*4]
    test eax, eax
    jz @f
    inc edx
    mov [edi], ecx
    mov [edi+4], eax
    add edi, 8
@@:
    inc ecx
    cmp ecx, 10h
    jne .init_next_leaf
    mov eax, edx
    retn

.grow_tree:
    pushad
    ;    
    lea esi, [leafs+4]
    lea edi, [nodes]
    xor ecx, ecx
@@:
    lea ebx, [ecx+ecx]
    mov eax, [leafs+ebx*4+4]
    mov [edi+ebx*8], eax
    inc ecx
    cmp ecx, [NumberOfLeafs]
    jnz @b

    dec ecx
    jz .exit_grow_tree	      ;     ,      ,
@@:			      ;  :)
    push ecx
    call .init_topnodes       ;   
    call .add_node	      ; 
    pop ecx
    loop @b
.exit_grow_tree:
    popad
    retn

.init_topnodes:
    lea edi, [topnodes]
    push edi
    xor eax, eax
    mov ecx, 16
    rep stosd
    pop edi

    xor ecx, ecx
@@:
    mov eax, 10h
    mul cl
    lea eax, [eax+nodes]
    push ecx
    call .get_topparent
    call .add_topnode
    pop ecx
    inc ecx
    cmp ecx, [NumberOfLeafs]
    jnz @b
    retn


.get_topparent:
    push eax
    mov eax, [eax+0Ch]
    test eax, eax	   ;parent ?
    jz @f		   ;,   
    call .get_topparent    ;, ,  parent' 
    add esp, 4
    retn
@@:
    pop eax
    retn

.add_topnode:
    lea edi, [topnodes]
    mov ecx, 11h
    repne scasd
    test ecx, ecx
    jz @f
    retn
@@:
    lea edi, [topnodes]
    mov ecx, 16
    push eax
    xor eax, eax
    repne scasd
    sub edi, 4
    pop dword[edi]
    retn

.add_node:
    call .get_lownodes	   ;     
			   ;eax -     
			   ;ebx -   , edi -   
    lea esi, [nodes]
    xor edx, edx
@@:
    cmp [esi], edx
    jz @f
    add esi, 10h
    jmp @b
@@:
    mov [esi], eax
    mov [esi+4], ebx
    mov [esi+8], edi
    mov [edi+0Ch], esi
    mov [ebx+0Ch], esi
    retn

.get_lownodes:
    mov ecx, 10h
    lea esi, [topnodes]
    or edx, -1
.srch_frst_lownode:
    dec ecx
    jz .get_scnd_lownode
    mov eax, [esi]
    test eax, eax
    jz .get_scnd_lownode
    mov eax, [eax]
    cmp edx, eax
    jb @f
    mov edx, eax
    mov ebx, [esi]
@@:
    add esi, 4
    jmp .srch_frst_lownode

.get_scnd_lownode:
    push edx
    or edx, -1
    mov ecx, 10h
    lea esi, [topnodes]
.srch_scnd_lownode:
    dec ecx
    jz .exit_get_lownodes
    mov eax, [esi]
    test eax, eax
    jz .exit_get_lownodes
    mov eax, [eax]
    cmp edx, eax
    jb @f
    cmp ebx, [esi]
    je @f
    mov edx, eax
    mov edi, [esi]
@@:
    add esi, 4
    jmp .srch_scnd_lownode
.exit_get_lownodes:
    pop eax
    add eax, edx
    retn

.init_udata:
    lea edi, [udata]
    mov ecx, [NumberOfLeafs]
    lea ecx, [ecx*2-2]
    lea ecx, [ecx+ecx]
    lea esi, [nodes+ecx*8]
    xor edx, edx
    xor ecx, ecx
    mov ebx, [esi+4h]	;lnode
    test ebx, ebx	 ;,        
    jz .only_one_leaf
    xor eax, eax
    call .bypass_tree
    mov ebx, [esi+8]
    mov al, 1
    call .bypass_tree
    retn
.only_one_leaf:
    mov eax, [leafs]
    mov byte[udata], al
    mov byte[udata+1], 1
    mov word[udata+2], 0
    retn

.bypass_tree:
    push edx
    push ebx
    push eax
    push ecx
    inc ecx
    shl edx, 1
    test eax, eax
    jz @f
    inc edx
@@:
    mov ebx, [ebx+4]
    test ebx, ebx
    jz @f
    xor eax, eax
    call .bypass_tree
@@:
    mov ebx, [esp+8]
    mov ebx, [ebx+8]
    test ebx, ebx
    jz @f
    mov al, 1
    call .bypass_tree
@@:
    mov ebx, [esp+8]
    cmp dword[ebx+4], 0
    jnz .exit_bypass_tree
    cmp dword[ebx+8], 0
    jnz .exit_bypass_tree
    push esi
    lea esi, [nodes]
    xor eax, eax
@@:
    cmp esi, ebx
    jz @f
    inc eax
    add esi, 10h
    jmp @b
@@:
    mov eax, [eax*8+leafs]
    pop esi
    mov byte[edi], al
    mov byte[edi+1], cl
    mov word[edi+2], dx
    add edi, 4
.exit_bypass_tree:
    pop ecx
    pop eax
    pop ebx
    pop edx
    retn

.calc_required_size:
    xor edx, edx
    xor ebx, ebx
    xor edi, edi
    xor esi, esi
    mov ecx, [NumberOfLeafs]
.calc_required_size_loop:
    sub ecx, 1
    jc .exit_calc_required_size
    movzx eax, byte[udata+ecx*4]
    mov bl, byte[udata+ecx*4+1]
    push ecx
    xor ecx, ecx
@@:
    cmp eax, [leafs+ecx*8]
    je @f
    inc ecx
    jmp @b
@@:
    mov eax, [leafs+ecx*8+4]
    mul ebx
    add edi, eax
    jnc @f
    inc esi
@@:
    pop ecx
    jmp .calc_required_size_loop
.exit_calc_required_size:
    mov [HNumberOfBits], esi
    mov [LNumberOfBits], edi
    mov eax, edi
    and eax, 7
    shr edi, 3
    test eax, eax
    jz @f
    inc edi
@@:
    rol esi, 3
    or edi, esi
    mov eax, edi
    retn

.pack_data:
    call .configure_udata
    mov esi, [lpSrc]
    mov edi, [lpRes]
    mov edx, [dwSizeOfSrc]

    mov ecx, [SizeOfRes]
    xor eax, eax
    push edi
    rep stosb
    pop edi

    mov dword[edi], 0
    mov eax, [dwSizeOfSrc]
    mov [edi+4], eax
    mov eax, [HNumberOfBits]
    mov [edi+8], eax
    mov eax, [LNumberOfBits]
    mov [edi+0Ch], eax

    add edi, 10h
    push esi
    mov ecx, 10h
    lea esi, [udata]
    rep movsd
    pop esi

    xor ebx, ebx
    xor ecx, ecx
.pack_next_byte:
    mov al, [esi]

    push esi
    mov bl, al
    shr bl, 4
    movzx esi, word[udata+ebx*4+2]
    mov cl, 20h
    sub cl, byte[udata+ebx*4+1]
    rol esi, cl
    xchg cl, ch
    ror esi, cl
    xchg cl, ch
    bswap esi
    or [edi], esi
    add ch, byte[udata+ebx*4+1]
    cmp ch, 10h
    jbe @f
    sub ch, 10h
    add edi, 2
@@:
    mov bl, al
    and bl, 0Fh
    movzx esi, word[udata+ebx*4+2]
    mov cl, 20h
    sub cl, byte[udata+ebx*4+1]
    rol esi, cl
    xchg cl, ch
    ror esi, cl
    xchg cl, ch
    bswap esi
    or [edi], esi
    add ch, byte[udata+ebx*4+1]
    cmp ch, 10h
    jbe @f
    sub ch, 10h
    add edi, 2
@@:
    pop esi
    inc esi
    dec edx
    jnz .pack_next_byte
    retn

.configure_udata:
    xor ebx, ebx
    lea esi, [udata]
    mov edi, esi
    sub esp, 40h

    push edi
    lea edi, [esp+4]
    mov ecx, 10h
    xor eax, eax
    rep stosd
    pop edi

    mov ecx, 0Fh
@@:
    mov eax, [esi+ecx*4]
    mov bl, al
    mov [esp+ebx*4], eax
    sub ecx, 1
    jnc @b
    mov esi, esp
    mov ecx, 10h
    rep movsd
    add esp, 40h
    retn
endp